/*
 * Copyright (c) 2014, 2015, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.oracle.truffle.r.runtime;

import com.oracle.truffle.api.*;
import com.oracle.truffle.api.frame.*;
import com.oracle.truffle.api.nodes.*;
import com.oracle.truffle.api.source.*;
import com.oracle.truffle.r.options.*;

/**
 * This class is intended to be used for internal errors that do not correspond to R errors.
 */
public final class RInternalError extends Error {

    private static final long serialVersionUID = 80698622974155216L;

    public RInternalError(String message, Object... args) {
        super(String.format(message, args));
        reportError(this, null);
    }

    public RInternalError(Throwable cause, String message, Object... args) {
        super(String.format(message, args), cause);
        reportError(this, null);
    }

    public static RuntimeException unimplemented() {
        CompilerDirectives.transferToInterpreter();
        throw new RInternalError("not implemented");
    }

    public static RuntimeException shouldNotReachHere() {
        CompilerDirectives.transferToInterpreter();
        throw new RInternalError("should not reach here");
    }

    public static RuntimeException shouldNotReachHere(Throwable cause) {
        CompilerDirectives.transferToInterpreter();
        throw new RInternalError(cause, "should not reach here");
    }

    public static RuntimeException shouldNotReachHere(String msg) {
        CompilerDirectives.transferToInterpreter();
        throw new RInternalError("should not reach here: %s", msg);
    }

    static void reportError(Throwable t, SourceSection source) {
        CompilerDirectives.transferToInterpreter();
        if (FastROptions.PrintErrorStacktraces.getValue()) {
            System.err.printf("RError with message %s:%n", t.getMessage());
            if (source != null) {
                System.err.printf("        at %s%n", source.getShortDescription());
            }
            Truffle.getRuntime().iterateFrames(frame -> {
                String sourceDesc = findSourceDesc(frame);
                System.err.printf("        at %s%n", sourceDesc != null ? sourceDesc : frame.getCallTarget().toString());
                return null;
            });
            System.err.println("Java stack trace:");
            t.printStackTrace();
        }
    }

    private static String findSourceDesc(FrameInstance frame) {
        if (frame.getCallNode() == null) {
            return null;
        }
        Node current = frame.getCallNode();
        do {
            if (current.getSourceSection() != null) {
                return current.getSourceSection().getShortDescription();
            }
            current = current.getParent();
        } while (current != null);
        return null;
    }
}
